"""This module implements the ISFED with denoising reconstruction method"""
import torch

from ..enhancing.spitfire_denoise import SpitfireDenoise

from .interface import SAiryscanReconstruction
from ._sure import SureMap
from .spitfire_join_denoise import SpitfireJoinDenoise


class ISFEDDenoising(SAiryscanReconstruction):
    """Reconstruct a high resolution image with a ISFED method including a denoising step

    The denoising is performed on the two terms of the ISFED image difference using the SPITFIR(e)
    algorithm

    :param epsilon: weighting parameter for the ISFED difference second term. If epsilon='map',
        epsilon is an automatic estimated weight map using the SURE criterion. If epsilon='mode',
        epsilon is a float which corresponds to the main mode of the SURE map. Otherwise, epsilon
        can be fixed to any float value
    :param reg_inner: Regularization for denoising of the first term of ISFED reconstruction
    :param reg_outer: Regularization for the denoising of the second term of the SFED reconstruction
    :param weighting_inner: Weighting parameter of the SPITFIR(e) denoising model for the first
        term of ISFED. Must be in [0, 1], with value close to 0 for sparse signal and close to
        one otherwise.
    :param weighting_outer: Weighting parameter of the SPITFIR(e) denoising model for the second
        term of ISFED. Must be in [0, 1], with value close to 0 for sparse signal and close to one
        otherwise.
    """
    def __init__(self,
                 epsilon: float = 0.3,
                 reg_inner: float = 0.995,
                 reg_outer: float = 0.995,
                 weighting_inner: float = 0.9,
                 weighting_outer: float = 0.9,
                 join_denoising: bool = False):
        super().__init__()
        self.num_args = 2
        self.epsilon = epsilon
        self.reg_inner = reg_inner
        self.reg_outer = reg_outer
        self.weighting_inner = weighting_inner
        self.weighting_outer = weighting_outer
        self.join_denoising = join_denoising
        self.map_ = None

    def __call__(self, image: torch.Tensor, reg_image: torch.Tensor) -> torch.Tensor:
        """Do the reconstruction

        :param image: Raw airyscan image. [H, Z, Y, X] for 3D image, [H, Y, X] for 2D images,
        :param reg_image: Co-registered airyscan image. [H, Z, Y, X] for 3D image, [H, Y, X] for
            2D images,
        :return: The reconstructed image. [Z, Y, X] for 3D, [Y, X] for 2D
        """
        if self.join_denoising:
            den_a_filter = SpitfireJoinDenoise(weight=self.weighting_inner, reg=self.reg_inner)
            a = 32*den_a_filter(reg_image).detach()
            den_b_filter = SpitfireJoinDenoise(weight=self.weighting_outer, reg=self.reg_outer)
            b = 32*den_b_filter(image).detach()
        else:
            a = torch.sum(reg_image, axis=0)
            b = torch.sum(image, axis=0)
            den_a_filter = SpitfireDenoise(weight=self.weighting_inner, reg=self.reg_inner)
            a = den_a_filter(a).detach()
            den_b_filter = SpitfireDenoise(weight=self.weighting_outer, reg=self.reg_outer)
            b = den_b_filter(b).detach()

        if self.epsilon == 'map':
            map_ = SureMap(smooth=False)
            epsilon = map_(image[0, ...], a, b)
            print('epsilon map=', epsilon.shape)
        elif self.epsilon == 'mode':
            map_ = SureMap(smooth=False)
            epsilon = map_.mode(image[0, ...], a, b)
            print('mode epsilon=', epsilon)
        else:
            epsilon = float(self.epsilon)
        out = a - epsilon * b
        return self._crop(torch.nn.functional.relu(out, inplace=True))


metadata = {
    'name': 'ISFEDDenoising',
    'label': 'ISFED Denoising',
    'class': ISFEDDenoising,
    'parameters': {
        'epsilon': {
            'type': str,
            'label': 'epsilon',
            'help': 'Weighting parameter',
            'default': 0.3
        },
        'reg_inner': {
            'type': float,
            'label': 'Inner reg',
            'help': 'Regularisation for inner detectors denoising',
            'default': 0.995
        },
        'reg_outer': {
            'type': float,
            'label': 'Outer reg',
            'help': 'Regularisation for outer detectors denoising',
            'default': 0.995
        },
        'weighting': {
            'type': float,
            'label': 'Sparsity weighting',
            'help': 'Denoising sparsity: 0.1 sparse, 0.6 moderately sparse, 0.9 not sparse',
            'default': 0.9,
            'range': (0, 1)
        }
    }
}



"""Module to define the processing class interface using the observer pattern"""
from abc import ABC, abstractmethod
import torch


class SObserver(ABC):
    """Interface of observer to notify progress

    An observer must implement the progress and message

    """
    def __init__(self):
        pass

    @abstractmethod
    def notify(self, message: str):
        """Notify a progress message

        :param message: Progress message
        """
        raise NotImplementedError('SObserver is abstract')

    @abstractmethod
    def progress(self, value: int):
        """Notify progress value

        :param value: Progress value in [0, 100]
        """
        raise NotImplementedError('SObserver is abstract')


class SObservable:
    """Interface for any data processing class

    The observable class can notify the observers for progress
    """
    def __init__(self):
        self._observers = []

    def add_observer(self, observer: SObserver):
        """Add an observer

        :param observer: Observer instance

        """
        self._observers.append(observer)

    def notify(self, message: str):
        """Notify progress to observers

        :param message: Progress message
        """
        for obs in self._observers:
            obs.notify(message)

    def progress(self, value: int):
        """Notify progress to observers

        :param value: Progress value in [0, 100]
        """
        for obs in self._observers:
            obs.progress(value)

    def __call__(self, image: torch.Tensor) -> torch.Tensor:
        """Run the data processing

        :param image: Image to process,
        :return: The processed image
        """


class SObserverConsole(SObserver):
    """Implementation of an observer that print messages and progress to console"""
    def notify(self, message):
        print(message)

    def progress(self, value):
        print('progress:', value, '%')



"""This module implements the reconstruction methods"""
from .interface import SAiryscanReconstruction
from .ifed import IFED
from .isfed import ISFED
from .ifed_denoising import IFEDDenoising
from .isfed_denoising import ISFEDDenoising
from .ism import ISM
from .pseudo_confocal import PseudoConfocal
from .spitfire_join_deconv import SpitfireJoinDeconv
from .spitfire_join_denoise import SpitfireJoinDenoise
from .spitfire_flow import SSpitfireFlow
from ._sure import SureMap


__all__ = [
    'SAiryscanReconstruction',
    'IFED',
    'ISFED',
    'ISM',
    'PseudoConfocal',
    'SpitfireJoinDeconv',
    'SpitfireJoinDenoise',
    'SSpitfireFlow',
    'SureMap',
    'IFEDDenoising',
    'ISFEDDenoising'
]



"""This module implements the IFED with denoising reconstruction method"""
import torch

from ..enhancing.spitfire_denoise import SpitfireDenoise

from .interface import SAiryscanReconstruction
from ._sure import SureMap
from .spitfire_join_denoise import SpitfireJoinDenoise


class IFEDDenoising(SAiryscanReconstruction):
    """Reconstruct a high resolution image with a ISFED method including a denoising step

    The denoising is performed on the two terms of the ISFED image difference using the SPITFIR(e)
    algorithm.

    :param inner_ring_index: Index of the last detector of the inner ring [7, 19],
    :param epsilon: weighting parameter for the ISFED difference second term. If epsilon='map',
        epsilon is an automatic estimated weight map using the SURE criterion. If epsilon='mode',
        epsilon is a float which corresponds to the main mode of the SURE map. Otherwise, epsilon
        can be fixed to any float value,
    :param reg_inner: Regularization for denoising of the first term of ISFED reconstruction,
    :param reg_outer: Regularization for the denoising of the second term of the SFED
        reconstruction,
    :param weighting: Weighting parameter of the SPITFIR(e) denoising model. Must be in [0, 1],
        with value close to 0 for sparse signal and close to one otherwise.

    """
    def __init__(self,
                 inner_ring_index: int = 7,
                 epsilon: str | float = 0.3,
                 reg_inner: float = 0.995,
                 reg_outer: float = 0.995,
                 weighting: float = 0.9,
                 join_denoising: bool = False):
        super().__init__()

        print('ifed constructor inner index=', inner_ring_index)
        print('ifed constructor epsilon=', epsilon)
        self.num_args = 1
        self.reg_inner = reg_inner
        self.reg_outer = reg_outer
        self.inner_ring_index = inner_ring_index
        self.epsilon = epsilon
        self.weighting = weighting
        self.join_denoising = join_denoising
        self.map_ = None
        if inner_ring_index not in [7, 19]:
            raise ValueError('Inner ring index must be in (7, 19)')

    def __call__(self, image: torch.Tensor) -> torch.Tensor:
        """Do the reconstruction

        :param image: Raw detector stack to reconstruct [H (Z) Y X]
        :return: high resolution image [(Z) Y X]
        """
        self.progress(0)
        self.notify('IFED: sum inner and outer detectors')
        if self.join_denoising:
            den_a_filter = SpitfireJoinDenoise(weight=self.weighting, reg=self.reg_inner)
            a = den_a_filter(image[0:self.inner_ring_index, ...]).detach()
            den_b_filter = SpitfireJoinDenoise(weight=self.weighting, reg=self.reg_outer)
            b = den_b_filter(image[self.inner_ring_index + 1:32, ...]).detach()

            a *= self.inner_ring_index
            b *= (32-self.inner_ring_index)
        else:
            a = torch.sum(image[0:self.inner_ring_index, ...], axis=0)
            b = torch.sum(image[self.inner_ring_index + 1:32, ...], axis=0)
            self.notify('IFED: denoise rings')
            den_a_filter = SpitfireDenoise(weight=self.weighting, reg=self.reg_inner)
            a = den_a_filter(a)
            den_b_filter = SpitfireDenoise(weight=self.weighting, reg=self.reg_outer)
            b = den_b_filter(b)

        self.progress(33)
        print('ifed epsilon=', self.epsilon)
        if self.epsilon == 'map':
            self.notify('IFED: compute epsilon map')
            map_ = SureMap(smooth=False)
            epsilon = map_(image[0, ...], a, b)
            self.map_ = epsilon
            print('epsilon map=', epsilon.shape)
        elif self.epsilon == 'mode':
            self.notify('IFED: estimate epsilon using sure')
            map_ = SureMap(smooth=False)
            epsilon = map_.mode(image[0, ...], a, b)
            self.map_ = epsilon
            print('mode epsilon=', epsilon)
        else:
            self.notify('IFED: use manual epsilon')
            epsilon = float(self.epsilon)
        self.progress(75)
        self.notify('IFED: do sum and relu')
        out = a - epsilon * b
        self.progress(100)
        return self._crop(torch.nn.functional.relu(out, inplace=True))


metadata = {
    'name': 'IFEDDenoising',
    'label': 'IFED Denoising',
    'class': IFEDDenoising,
    'parameters': {
        'inner_ring_index': {
            'type': int,
            'label': 'Inner index',
            'help': 'Index of the inner ring last detector (7, 19)',
            'default': 7
        },
        'epsilon': {
            'type': str,
            'label': 'epsilon',
            'help': 'Weighting parameter',
            'default': 0.3
        },
        'reg_inner': {
            'type': float,
            'label': 'Inner reg',
            'help': 'Regularisation for inner detectors denoising',
            'default': 0.995
        },
        'reg_outer': {
            'type': float,
            'label': 'Outer reg',
            'help': 'Regularisation for outer detectors denoising',
            'default': 0.995
        },
        'weighting': {
            'type': float,
            'label': 'Sparsity weighting',
            'help': 'Denoising sparsity: 0.1 sparse, 0.6 moderately sparse, 0.9 not sparse',
            'default': 0.9,
            'range': (0, 1)
        }
    }
}



"""This module implements the Wiener deconvolution algorithm"""
import torch
from .interface import SAiryscanEnhancing


class SAiryscanWiener(SAiryscanEnhancing):
    """Apply a gaussian filter

    :param psf: Point spread function
    :param beta: Regularization parameter

    """
    def __init__(self, psf: torch.Tensor, beta: float = 1e-5):
        super().__init__()
        self.psf = psf
        self.beta = beta

    @staticmethod
    def _resize_psf(image: torch.Tensor, psf: torch.Tensor) -> torch.Tensor:
        """Pad a PSF image to match a given shape

        :param image: Target image shape
        :param psf: The Point Spread Function to resize,
        :return: The resized PSF image
        """
        kernel = torch.zeros(image.shape)
        x_start = int(image.shape[0] / 2 - psf.shape[0] / 2) + 1
        y_start = int(image.shape[1] / 2 - psf.shape[1] / 2) + 1
        kernel[x_start:x_start+psf.shape[0], y_start:y_start+psf.shape[1]] = psf
        return kernel

    def __call__(self, image: torch.Tensor) -> torch.Tensor:
        """Do the deconvolution

        :param image: Image to deblur [Y, X] or [Z, Y, X]
        :return: THe deblurred image
        """
        if image.ndim == 2:
            padding = 13
            pad_fn = torch.nn.ReflectionPad2d(padding)
            image_pad = pad_fn(image.detach().clone().view(1, 1,
                image.shape[0], image.shape[1])).view(
                    (image.shape[0] + 2 * padding, image.shape[1] + 2 * padding))
            fft_source = torch.fft.fft2(image_pad)
            psf = self._resize_psf(image_pad, self.psf)
            psf_roll = torch.roll(psf, int(-psf.shape[0] / 2), dims=0)
            psf_roll = torch.roll(psf_roll, int(-psf.shape[1] / 2), dims=1)
            fft_psf = torch.fft.fft2(psf_roll)
            return torch.real(torch.fft.ifft2(fft_source * torch.conj(fft_psf) / (
                 self.beta**2 + fft_psf * torch.conj(fft_psf))))[padding:-padding, padding:-padding]
        if image.ndim == 3:
            fft_source = torch.fft.fftn(image)
            psf_roll = torch.roll(self.psf, int(-self.psf.shape[0] / 2), dims=0)
            psf_roll = torch.roll(psf_roll, int(-self.psf.shape[1] / 2), dims=1)
            psf_roll = torch.roll(psf_roll, int(-self.psf.shape[2] / 2), dims=2)
            fft_psf = torch.fft.fftn(psf_roll)
            return torch.real(torch.fft.ifftn(fft_source * torch.conj(fft_psf) / (
                self.beta * fft_source * torch.conj(fft_source) + fft_psf * torch.conj(fft_psf))))
        raise ValueError("The image shape does not match for 2D or 3D images")


metadata = {
    'name': 'SAiryscanWiener',
    'label': 'Wiener',
    'class': SAiryscanWiener,
    'parameters': {
        'psf': {
            'type': torch.Tensor,
            'label': 'psf',
            'help': 'Point Spread Function',
            'default': None
        },
        'beta': {
            'type': float,
            'label': 'Beta',
            'help': 'Regularisation parameter',
            'default': 1e-5,
            'range': (0, 999999)
        }
    }
}



"""This module implement the SURE estimation of weights between two images"""
import numpy as np
import torch
from torchvision import transforms
from skimage.filters import threshold_otsu


class SureMap:
    """Calculate the Sure weights map for IFED and ISFED

    :param smooth: True to smooth the map with a sigma=2 gaussian filter
    """
    def __init__(self, smooth: float = False):
        self.patch_size = 5
        self.smooth = smooth

    def __call__(self,
                 image_ref: torch.Tensor,
                 image_a: torch.Tensor,
                 image_b: torch.Tensor
                 ) -> torch.Tensor:
        """Do the calculation

        :param image_ref: Reference image
        :param image_a: First image to combine
        :param image_b: Second image to combine
        :return: The sure map, where each coefficient the weight between the two images
        """
        width = image_ref.shape[0]
        height = image_ref.shape[1]
        image_ref = image_ref.view(1, 1, image_ref.shape[0], image_ref.shape[1])
        image_a = image_a.view(1, 1, image_a.shape[0], image_a.shape[1])
        image_b = image_b.view(1, 1, image_b.shape[0], image_b.shape[1])

        # extract patch
        image_ref_patch = torch.nn.functional.unfold(image_ref, self.patch_size,
                                                     padding=self.patch_size//2)
        image_a_patch = torch.nn.functional.unfold(image_a, self.patch_size,
                                                   padding=self.patch_size//2)
        image_b_patch = torch.nn.functional.unfold(image_b, self.patch_size,
                                                   padding=self.patch_size//2)

        # sure
        num = torch.sum((image_ref_patch-image_a_patch)*image_b_patch, dim=1)
        den = torch.sum(image_b_patch*image_b_patch, dim=1)
        sure_map = -num/den
        sure_map = sure_map.view(width, height)

        if self.smooth:
            return transforms.functional.gaussian_blur(sure_map.view(1, 1, width, height),
                                                       kernel_size=(7, 7),
                                                       sigma=2).view(width, height)
        return sure_map

    def mode(self,
             image_ref: torch.Tensor,
             image_a: torch.Tensor,
             image_b: torch.Tensor
             ) -> float:
        """Calculate a single weight coefficient to combine image_a and image_b

        :param image_ref: Reference image
        :param image_a: First image to combine
        :param image_b: Second image to combine
        :return: The weighting coefficient
        """
        sure_map = self(image_ref, image_a, image_b).detach().numpy()
        th = threshold_otsu(sure_map)
        return np.mean(sure_map[sure_map >= th])



"""This module implements the command line interface to run SAiryscan reconstructions"""
import argparse
from skimage.io import imsave

from ..api import SAiryscanAPI


def add_args_to_parser(parser: argparse.ArgumentParser, api: SAiryscanAPI):
    """Add arguments to the command line parser from the API

    :param parser: Parser to modify
    :param api: API that contains all the available parameters
    """
    for filter_name in api.filters.get_keys():
        params = api.filters.get_parameters(filter_name)
        for key, value in params.items():
            parser.add_argument(f"--{key}", help=value['help'], default=value['default'])


def main():
    """Entry point for the command line tool"""
    parser = argparse.ArgumentParser(description='SAiryscan reconstruction',
                                     conflict_handler='resolve')

    parser.add_argument('-i', '--input', help='Input image file', default='.czi')
    parser.add_argument('-r', '--reg', help='Registration method', default='none')
    parser.add_argument('-m', '--method', help='Reconstruction method', default='ISM')
    parser.add_argument('-e', '--enhance', help='Post processing enhancing', default='none')
    parser.add_argument('-o', '--output', help='Output image file', default='.tif')

    api = SAiryscanAPI()
    add_args_to_parser(parser, api)
    args = parser.parse_args()

    args_dict = vars(args)
    print('args=', vars(args))

    # instantiate pipeline
    if args.reg != 'none':
        reg_filter = api.filter(args.reg, **args_dict)
    else:
        reg_filter = None
    if args.method != 'none':
        rec_filter = api.filter(args.method, **args_dict)
    else:
        rec_filter = None
    if args.enhance != 'none':
        enhance_filter = api.filter(args.enhance, **args_dict)
    else:
        enhance_filter = None
    pipeline = api.pipeline(rec_filter, reg_filter, enhance_filter)

    # run pipeline
    reader = api.reader(args.input)
    out_image = pipeline(reader.data())
    imsave(args.output, out_image.detach().numpy())


if __name__ == "__main__":
    main()



"""This module implement the SAiryscan main API"""
import os
import importlib

from ..core import SAiryscanReader, SObservable
from ..registration import SAiryscanRegistration
from ..reconstruction import SAiryscanReconstruction
from ..enhancing import SAiryscanEnhancing

from .factory import SAiryscanModuleFactory
from .pipeline import SAiryscanPipeline, SAiryscanLoop


class SAiryscanAPI:
    """Main API to build an Airyscan data reconstruction pipeline"""
    def __init__(self):
        self.filters = SAiryscanModuleFactory()
        discovered_modules = self._find_modules()
        for name in discovered_modules:
            # print('register the module:', name)
            mod = importlib.import_module(name)
            # print(mod.__name__)
            self.filters.register(mod.metadata['name'], mod.metadata)

    @staticmethod
    def _find_modules() -> list[str]:
        """Search for modules to instantiate in the factory

        :return: The list of module found
        """
        path = os.path.abspath(os.path.dirname(__file__))
        path = os.path.dirname(path)
        modules = []
        for parent in ['enhancing', 'reconstruction', 'registration']:
            path_ = os.path.join(path, parent)
            for x in os.listdir(path_):
                if (x.endswith(".py") and 'interface' not in x
                        and '__init__' not in x and not x.startswith("_")):
                    modules.append(f"sairyscan.{parent}.{x.split('.')[0]}")
        return modules

    def filter(self, name: str, **args) -> SObservable | None:
        """Get a filter

        :param name: Unique name of the filter
        :return: An instance of the filter if it exists, None otherwise
        """
        if name == 'None':
            return None
        return self.filters.get(name, **args)

    @staticmethod
    def reader(filename: str) -> SAiryscanReader:
        """Get the airyscan image file reader

        :param filename: Path of the image file
        :return: An intance of the reader
        """
        return SAiryscanReader(filename)

    @staticmethod
    def pipeline(reconstruction: SAiryscanReconstruction,
                 registration: SAiryscanRegistration = None,
                 enhancing: SAiryscanEnhancing = None
                 ) -> SAiryscanPipeline:
        """Get a reconstruction pipeline

        :param reconstruction: instance of the reconstruction filter
        :param registration: instance of the registration filter
        :param enhancing: instance of the enhancing filter
        :return: An instance of the pipeline
        """
        return SAiryscanPipeline(reconstruction, registration, enhancing)

    @staticmethod
    def loop(method: SObservable,
             filename: str,
             to_file: bool = False,
             destination_filename: str = ''
             ) -> SAiryscanLoop:
        """Get an instance of pipeline loop

        :param method: Pipeline or filter to apply,
        :param filename: Path to the file containing the data,
        :param to_file: True to dave the tensor to file. Otherwise, the call method return a tensor
        :param destination_filename: Path to the file where the result is saved
        :return: An instance of the loop
        """
        instance_ = SAiryscanLoop(filename, to_file, destination_filename)
        return instance_(method)



"""Implementation of registration with fixed position"""
import torch
from torchvision import transforms
from .interface import SAiryscanRegistration


class SRegisterPosition(SAiryscanRegistration):
    """Register the detectors stack by translating each image to the detector position is array

    :param weight: Weight applied on the detector position translation

    """
    def __init__(self, weight: float = 1.0):
        super().__init__()
        self.weight = weight

    def __call__(self, image: torch.Tensor) -> torch.Tensor:
        """Run registration

        :param image: Stack to register
        :return: The registered stack
        """
        image_out = torch.zeros(image.shape, dtype=torch.float32)
        d = self.weight
        image_out[0, ...] = image[0, ...].clone()
        image_out[1, ...] = self._translate_detector(image[1, ...], (d, 0.5 * d))
        image_out[2, ...] = self._translate_detector(image[2, ...], (d, -0.5 * d))
        image_out[3, ...] = self._translate_detector(image[3, ...], (0, -d))
        image_out[4, ...] = self._translate_detector(image[4, ...], (-d, -0.5 * d))
        image_out[5, ...] = self._translate_detector(image[5, ...], (-d, 0.5 * d))
        image_out[6, ...] = self._translate_detector(image[6, ...], (0, d))
        image_out[7, ...] = self._translate_detector(image[7, ...], (d, 1.5 * d))
        image_out[8, ...] = self._translate_detector(image[8, ...], (2 * d, d))
        image_out[9, ...] = self._translate_detector(image[9, ...], (2 * d, 0))
        image_out[10, ...] = self._translate_detector(image[10, ...], (2 * d, -d))
        image_out[11, ...] = self._translate_detector(image[11, ...], (d, -1.5 * d))
        image_out[12, ...] = self._translate_detector(image[12, ...], (0, -2 * d))
        image_out[13, ...] = self._translate_detector(image[13, ...], (-d, -1.5 * d))
        image_out[14, ...] = self._translate_detector(image[14, ...], (-2 * d, -d))
        image_out[15, ...] = self._translate_detector(image[15, ...], (-2 * d, 0))
        image_out[16, ...] = self._translate_detector(image[16, ...], (-2 * d, d))
        image_out[17, ...] = self._translate_detector(image[17, ...], (-d, 1.5 * d))
        image_out[18, ...] = self._translate_detector(image[18, ...], (0, 2 * d))
        image_out[19, ...] = self._translate_detector(image[19, ...], (d, 2.5 * d))
        image_out[20, ...] = self._translate_detector(image[20, ...], (2 * d, 2 * d))
        image_out[21, ...] = self._translate_detector(image[21, ...], (3 * d, 0.5 * d))
        image_out[22, ...] = self._translate_detector(image[22, ...], (3 * d, -0.5 * d))
        image_out[23, ...] = self._translate_detector(image[23, ...], (2 * d, -2 * d))
        image_out[24, ...] = self._translate_detector(image[24, ...], (d, -2.5 * d))
        image_out[25, ...] = self._translate_detector(image[25, ...], (-d, -2.5 * d))
        image_out[26, ...] = self._translate_detector(image[26, ...], (-2 * d, -2 * d))
        image_out[27, ...] = self._translate_detector(image[27, ...], (-3 * d, -0.5 * d))
        image_out[28, ...] = self._translate_detector(image[28, ...], (-3 * d, 0.5 * d))
        image_out[29, ...] = self._translate_detector(image[29, ...], (-2 * d, 2 * d))
        image_out[30, ...] = self._translate_detector(image[30, ...], (-1 * d, 2.5 * d))
        image_out[31, ...] = self._translate_detector(image[31, ...], (0, 3 * d))
        return image_out

    @staticmethod
    def _translate_detector(img: torch.Tensor, translate: tuple[float, float]):
        """Translate one detector

        :param img: Tensor for one single detector
        :param translate: (x, y, z) or (x, y) translation vector
        :return: the translated tensor
        """
        img_bc = img.view((1, 1, img.shape[0], img.shape[1]))
        img_out = transforms.functional.affine(img_bc, angle=0, translate=translate, scale=1,
                                    shear=[0, 0],
                                    interpolation=transforms.functional.InterpolationMode.BILINEAR)
        return img_out.view(img.shape)


metadata = {
    'name': 'SRegisterPosition',
    'label': 'Detector position',
    'class': SRegisterPosition,
    'parameters': {
        'weight': {
            'type': float,
            'label': 'weight',
            'help': 'Translation weight',
            'default': 1,
            'range': [-10, 10]
        }
    }
}



"""This module contain all the API to interact with the SAiryscan reconstruction modules"""
from .api import SAiryscanAPI
from .pipeline import SAiryscanPipeline
from .pipeline import SAiryscanLoop

__all__ = ['SAiryscanAPI',
           'SAiryscanPipeline',
           'SAiryscanLoop']



"""Interface for Airyscan registration filter"""
import torch
from ..core import SObservable


class SAiryscanRegistration(SObservable):
    """Interface for Airyscan detectors registration method"""

    def __call__(self, image: torch.Tensor) -> torch.Tensor:
        """Do the registration

        :param image: Raw airyscan data for a single channel time point [H (Z) Y X]
        :return: co-registered detectors [H (Z) Y X]
        """
        raise NotImplementedError('SairyscanRegistration is an interface. Please implement the'
                                  ' __call__ method')



"""This module implements the ISFED reconstruction method"""
import torch
from .interface import SAiryscanReconstruction
from ._sure import SureMap


class ISFED(SAiryscanReconstruction):
    """Implementation of the ISFED reconstruction method

    :param epsilon: Weight between the two images to combine
    """
    def __init__(self, epsilon=0.3):
        super().__init__()
        self.num_args = 2
        self.epsilon = epsilon

    def __call__(self, image: torch.Tensor, reg_image: torch.Tensor) -> torch.Tensor:
        """Do the reconstruction

        :param image: Raw detector stack to reconstruct [H (Z) Y X]
        :param reg_image: Spatially co-registered detectors stack [H (Z) Y X]
        :return: high resolution image [(Z) Y X]
        """
        a = torch.sum(reg_image, axis=0)
        b = torch.sum(image, axis=0)

        if self.epsilon == 'map':
            map_ = SureMap(smooth=True)
            epsilon = map_(image[0, ...], a, b)
            print('epsilon map=', epsilon.shape)
        elif self.epsilon == 'mode':
            map_ = SureMap(smooth=False)
            epsilon = map_.mode(image[0, ...], a, b)
            print('mode epsilon=', epsilon)
        else:
            epsilon = self.epsilon
        out = a - epsilon * b
        return self._crop(torch.nn.functional.relu(out, inplace=True))


metadata = {
    'name': 'ISFED',
    'label': 'ISFED',
    'class': ISFED,
    'parameters': {
        'epsilon': {
            'type': str,
            'label': 'epsilon',
            'help': 'Weighting parameter',
            'default': 0.3
        }
    }
}



"""This module implements the Richardson-Lucy deconvolution algorithm"""
import torch
from .interface import SAiryscanEnhancing


class SAiryscanRichardsonLucy(SAiryscanEnhancing):
    """Apply a Richardson-Lucy deconvolution

    :param psf: Point spread function
    :param niter: Number of iterations
    """
    def __init__(self, psf: torch.Tensor, niter: int = 30):
        super().__init__()
        self.psf = psf
        self.niter = niter

    @staticmethod
    def _resize_psf(psf: torch.Tensor, width: int, height: int) -> torch.Tensor:
        """Pad a PSF image to match a given shape

        :param psf: The Point Spread Function to resize,
        :param width: Targeted image width,
        :param height: Targeted image height,
        :return: The resized PSF image
        """
        kernel = torch.zeros((width, height))
        x_start = int(width / 2 - psf.shape[0] / 2) + 1
        y_start = int(height / 2 - psf.shape[1] / 2) + 1
        kernel[x_start:x_start+psf.shape[0], y_start:y_start+psf.shape[1]] = psf
        return kernel

    def __call__(self, image: torch.Tensor) -> torch.Tensor:
        """Do the deconvolution

        :param image: Image to deblur [Y, X] or [Z, Y, X]
        :return: The deblurred image
        """
        if image.ndim == 2:
            padding = 13
            psf = self._resize_psf(self.psf, image.shape[0]+2*padding, image.shape[1]+2*padding)
            psf_roll = torch.roll(psf, [int(-psf.shape[0]/2),
                                        int(-psf.shape[1]/2)], dims=(0, 1))
            fft_psf = torch.fft.fft2(psf_roll)
            fft_psf_mirror = torch.fft.fft2(torch.flip(psf_roll, dims=[0, 1]))

            pad_fn = torch.nn.ReflectionPad2d(padding)
            image_pad = pad_fn(image.detach().clone().view(1, 1,
                            image.shape[0], image.shape[1])).view((image.shape[0]+2*padding,
                                                                   image.shape[1]+2*padding))
            out_image = pad_fn(image.detach().clone().view(1, 1,
                            image.shape[0], image.shape[1])).view((image.shape[0]+2*padding,
                                                                   image.shape[1]+2*padding))
            for _ in range(self.niter):
                fft_out = torch.fft.fft2(out_image)
                fft_tmp = fft_out*fft_psf
                tmp = torch.real(torch.fft.ifft2(fft_tmp))
                tmp = image_pad/tmp
                fft_tmp = torch.fft.fft2(tmp)
                fft_tmp = fft_tmp * fft_psf_mirror
                tmp = torch.real(torch.fft.ifft2(fft_tmp))
                out_image = out_image * tmp
            return out_image[padding:-padding, padding:-padding]

        raise NotImplementedError('Richardson Lucy 3D deconvolution is not yet implemented')


metadata = {
    'name': 'SAiryscanRichardsonLucy',
    'label': 'Richardson-Lucy',
    'class': SAiryscanRichardsonLucy,
    'parameters': {
        'psf': {
            'type': torch.Tensor,
            'label': 'psf',
            'help': 'Point Spread Function',
            'default': None
        },
        'niter': {
            'type': int,
            'label': 'niter',
            'help': 'Number of iterations',
            'default': 30,
            'range': (0, 999999)
        }
    }
}



"""This module gives access to example data for testing and demo SAiryscan"""
import os.path as osp
import os
import numpy as np
import torch

from sairyscan.core import SAiryscanReader

__all__ = ['celegans']

legacy_data_dir = osp.abspath(osp.dirname(__file__))


def _fetch(data_filename: str):
    """Fetch a given data file from the local data dir.
    This function provides the path location of the data file given
    its name in the scikit-image repository.

    :param data_filename: Name of the file in the library data dir,
    :return: Path of the local file as a python string.
    """

    filepath = os.path.join(legacy_data_dir, data_filename)

    if os.path.isfile(filepath):
        return filepath
    raise FileExistsError("Cannot find the file:", filepath)


def _load(f: str) -> np.ndarray:
    """Load an image file located in the data directory.

    :param f : File name.
    :return img : Image loaded from ``skimage.data_dir``.
    """
    # importing io is quite slow since it scans all the backends
    # we lazy import it here
    from skimage.io import imread
    return imread(_fetch(f))


def celegans() -> torch.Tensor:
    """2D Celegans sample expressing the ERM-1::mNeonGreen fusion protein.

    :return: celegans : (32, 316, 316) float ndarray
    """

    reader = SAiryscanReader(_fetch("celegans_airyscan.czi"))
    return reader.data()



"""This module implements strategies to weight the detectors"""
import math
import torch


class SAiryscanWeights:
    """Creates weight for each detector depending on a method

    :param method: Method to calculate weights:
                   ['mean', 'ring', 'ring_inv', 'd2c', 'id2c', 'exp_d2c', 'exp_d2c_inv']
    """
    def __init__(self, method: str = 'mean'):
        self.method = method

    def __call__(self) -> torch.Tensor:
        """Compute the weights

        :return: The weights array (len=32)
        """
        if self.method == 'mean':
            weights = torch.ones((32,))
            return weights / torch.sum(weights)
        if self.method == 'ring':
            return self.ring_distance(3.0)
        if self.method == 'ring_inv':
            return self.ring_inv_distance(3.0)
        if self.method == 'd2c':
            d2c = self.distance_to_center(1)
            return d2c / torch.sum(d2c)
        if self.method == 'id2c':
            id2c = 3.0 - self.distance_to_center(1)
            return id2c / torch.sum(id2c)
        if self.method == 'exp_d2c':
            return self.exp_distance_to_center(3.0)
        if self.method == 'exp_d2c_inv':
            return self.exp_inv_distance_to_center(3.0)
        raise ValueError(f'Not recognized weighting method: {self.method}')

    @staticmethod
    def ring_distance(tau: float) -> torch.Tensor:
        """Compute the weights as the distance of the detector ring to the center

        The distance is calculated as exp(d/tau), where d is the ring index and tau a parameter

        :param tau: Exponential factor to apply to the distance
        :return: An array with the weight of each detector
        """
        weights = torch.ones((32,))
        for i in range(1, 7):
            weights[i] = math.exp(1/tau)
        for i in range(7, 19):
            weights[i] = math.exp(2 / tau)
        for i in range(19, 32):
            weights[i] = math.exp(3 / tau)
        return weights/torch.sum(weights)

    @staticmethod
    def ring_inv_distance(tau: float) -> torch.Tensor:
        """Compute the weights as the inverted distance of the detector ring to the center

        The distance is calculated as exp(-d/tau), where d is the ring index and tau a parameter

        :param tau: Exponential factor to apply to the distance
        :return: An array with the weight of each detector
        """
        weights = torch.ones((32,))
        for i in range(1, 7):
            weights[i] = math.exp(-1 / tau)
        for i in range(7, 19):
            weights[i] = math.exp(-2 / tau)
        for i in range(19, 32):
            weights[i] = math.exp(-3 / tau)
        return weights/torch.sum(weights)

    @staticmethod
    def exp_distance_to_center(tau: float) -> torch.Tensor:
        """Compute the weights as the distance of the detector to the central detector

        The distance is calculated as exp(d/tau), where d is the distance to center and tau a
        parameter

        :param tau: Exponential factor to apply to the distance
        :return: An array with the weight of each detector
        """
        weights = SAiryscanWeights.distance_to_center(1)
        for i in range(32):
            weights[i] = math.exp(weights[i] / tau)
        return weights/torch.sum(weights)

    @staticmethod
    def exp_inv_distance_to_center(tau: float) -> torch.Tensor:
        """Compute the weights as the inverted distance of the detector to the central detector

        The distance is calculated as exp(-d/tau), where d is the distance to center and tau a
        parameter

        :param tau: Exponential factor to apply to the distance
        :return: An array with the weight of each detector
        """
        weights = SAiryscanWeights.distance_to_center(1)
        for i in range(32):
            weights[i] = math.exp(-weights[i] / tau)
        return weights/torch.sum(weights)

    @staticmethod
    def distance_to_center(d: float) -> torch.Tensor:
        """Compute the distance of each detector to the central detector

        :param d: Width of a detector
        :return: An array with the distance of each detector
        """
        dist = torch.zeros((32,))
        dist[0] = SAiryscanWeights.norm(0, 0)
        dist[1] = SAiryscanWeights.norm(d, 0.5*d)
        dist[2] = SAiryscanWeights.norm(d, -0.5*d)
        dist[3] = SAiryscanWeights.norm(0, -d)
        dist[4] = SAiryscanWeights.norm(-d, -0.5*d)
        dist[5] = SAiryscanWeights.norm(-d, 0.5*d)
        dist[6] = SAiryscanWeights.norm(0, d)
        dist[7] = SAiryscanWeights.norm(d, 1.5*d)
        dist[8] = SAiryscanWeights.norm(2*d, d)
        dist[9] = SAiryscanWeights.norm(2*d, 0)
        dist[10] = SAiryscanWeights.norm(2*d, -d)
        dist[11] = SAiryscanWeights.norm(d, -1.5*d)
        dist[12] = SAiryscanWeights.norm(0, -2*d)
        dist[13] = SAiryscanWeights.norm(-d, -1.5*d)
        dist[14] = SAiryscanWeights.norm(-2*d, -d)
        dist[15] = SAiryscanWeights.norm(-2*d, 0)
        dist[16] = SAiryscanWeights.norm(-2*d, d)
        dist[17] = SAiryscanWeights.norm(-d, 1.5*d)
        dist[18] = SAiryscanWeights.norm(0, 2*d)
        dist[19] = SAiryscanWeights.norm(d, 2.5*d)
        dist[20] = SAiryscanWeights.norm(2*d, 2*d)
        dist[21] = SAiryscanWeights.norm(3*d, 0.5*d)
        dist[22] = SAiryscanWeights.norm(3*d, -0.5*d)
        dist[23] = SAiryscanWeights.norm(2*d, -2*d)
        dist[24] = SAiryscanWeights.norm(d, -2.5*d)
        dist[25] = SAiryscanWeights.norm(-d, -2.5*d)
        dist[26] = SAiryscanWeights.norm(-2*d, -2*d)
        dist[27] = SAiryscanWeights.norm(-3*d, -0.5*d)
        dist[28] = SAiryscanWeights.norm(-3*d, 0.5*d)
        dist[29] = SAiryscanWeights.norm(-2*d, 2*d)
        dist[30] = SAiryscanWeights.norm(-1*d, 2.5*d)
        dist[31] = SAiryscanWeights.norm(0, 3*d)
        return dist

    @staticmethod
    def norm(a: float, b: float) -> float:
        """Compute the norm of a 2D vector

        :param a: First component of the vector,
        :param b: Second component of the vector,
        :return: The vector norm
        """
        return math.sqrt(a*a + b*b)



"""This module implements a pipeline maker tool for SAiryscan"""
import torch
from sairyscan.core.io import SAiryscanReader
from sairyscan.core import SObservable
from sairyscan.registration import SAiryscanRegistration
from sairyscan.reconstruction import SAiryscanReconstruction
from sairyscan.enhancing import SAiryscanEnhancing


class SAiryscanPipeline(SObservable):
    """Build and run a reconstruction pipeline in a single image

    :param reconstruction: instance of the reconstruction filter
    :param registration: instance of the registration filter
    :param enhancing: instance of the enhancing filter
    """
    def __init__(self,
                 reconstruction: SAiryscanReconstruction,
                 registration: SAiryscanRegistration = None,
                 enhancing: SAiryscanEnhancing = None):
        super().__init__()
        self.reconstruction = reconstruction
        self.registration = registration
        self.enhancing = enhancing

    def _transmit_observers(self, child: SObservable):
        """Transmit observer to the observable

        :param child: Object that will receive the observers
        """
        for observer in self._observers:
            child.add_observer(observer)

    def __call__(self, image: torch.Tensor):
        """Run the pipeline

        :param image: Raw airyscan image. [H, Z, Y, X] for 3D image, [H, Y, X] for 2D images,
        :return: The reconstructed image. [Z, Y, X] for 3D, [Y, X] for 2D
        """
        self._transmit_observers(self.reconstruction)
        if self.registration:
            self._transmit_observers(self.registration)
            reg_image = self.registration(image)
            if self.reconstruction.num_args == 2:
                rec_image = self.reconstruction(image, reg_image)
            else:
                rec_image = self.reconstruction(reg_image)
        else:
            rec_image = self.reconstruction(image)

        if self.enhancing:
            self._transmit_observers(self.enhancing)
            return self.enhancing(rec_image)
        return rec_image


class SAiryscanLoop(SObservable):
    """Apply a SAiryscan reconstruction method on all the frames and channels

    :param filename: Path to the file containing the data,
    :param to_file: True to dave the tensor to file. Otherwise, the call method return a tensor
    :param destination_filename: Path to the file where the result is saved
    """
    def __init__(self, filename: str, to_file=False, destination_filename=''):
        super().__init__()
        self.filename = filename
        self.to_file = to_file
        self.destination_filename = destination_filename
        self.reader = SAiryscanReader(filename)

    def __call__(self, method: SObservable) -> torch.Tensor | None:
        """Apply the reconstruction loop

        :param method: instance of airyscan filter or pipeline
        :return: a torch.Tensor if the input to_file is false
        """
        if self.to_file:
            return self.reconstruct_to_file(method)
        return self.reconstruct_to_tensor(method)

    def reconstruct_to_tensor(self, method: SObservable) -> torch.Tensor:
        """Implement the reconstruction to a tensor

        :param method: instance of airyscan filter or pipeline
        :return: a torch.Tensor with the processed image
        """
        if self.reader.depth() > 1:
            out_data = torch.zeros((self.reader.frames(), int(self.reader.channels()/2),
                                    self.reader.depth(),
                                    self.reader.height(), self.reader.width()))
        else:
            out_data = torch.zeros((self.reader.frames(), int(self.reader.channels()/2),
                                    self.reader.height(), self.reader.width()))
        for frame_idx in range(self.reader.frames()):
            for channel_idx in range(0, self.reader.channels(), 2):
                raw_data = self.reader.data(frame_idx, channel_idx)
                out_data[frame_idx, channel_idx, ...] = method(raw_data)
        return out_data.squeeze()

    def reconstruct_to_file(self, method: SObservable):
        """Reconstruct an image directly into a file (for large memory data)

        TODO: Implement this method

        :param method: instance of airyscan filter or pipeline
        """
        raise NotImplementedError('SAiryscanLoop: reconstruct_to_file() not yet implemented')



"""This module implements the Spitfire join deconvolution reconstruction method"""
import torch
from .interface import SAiryscanReconstruction
from ._detectors_weights import SAiryscanWeights


def hv_loss(img: torch.Tensor, weighting: float) -> torch.Tensor:
    """Sparse Hessian regularization term (or loss)

    :param img: Tensor of shape BCYX containing the estimated image
    :param weighting: Sparse weighting parameter in [0, 1]. 0 sparse, and 1 not sparse
    :return: The loss value
    """
    a, b, h, w = img.size()
    dxx2 = torch.square(-img[:, :, 2:, 1:-1] + 2 * img[:, :, 1:-1, 1:-1] - img[:, :, :-2, 1:-1])
    dyy2 = torch.square(-img[:, :, 1:-1, 2:] + 2 * img[:, :, 1:-1, 1:-1] - img[:, :, 1:-1, :-2])
    dxy2 = torch.square(img[:, :, 2:, 2:] - img[:, :, 2:, 1:-1] - img[:, :, 1:-1, 2:] +
                        img[:, :, 1:-1, 1:-1])
    hv = torch.sqrt(weighting * weighting * (dxx2 + dyy2 + 2 * dxy2) +
                    (1 - weighting) * (1 - weighting) * torch.square(img[:, :, 1:-1, 1:-1])).sum()
    return hv / (a * b * h * w)


def dataterm_deconv(blurry_image: torch.Tensor,
                    deblurred_image: torch.Tensor,
                    psf: torch.Tensor,
                    detector_weights: torch.Tensor
                    ) -> torch.Tensor:
    """Deconvolution L2 data-term

    Compute the L2 error between the original image and the convoluted reconstructed image

    :param blurry_image: Tensor of shape BCHYX containing the original blurry image
    :param deblurred_image: Tensor of shape BCYX containing the estimated deblurred image
    :param psf:  containing the point spread function
    :param detector_weights: Weight applied to each detector
    """
    a, b, h, w = deblurred_image.size()
    conv_op = torch.nn.Conv2d(1, 1, kernel_size=psf.shape[2],
                              stride=1,
                              padding=int((psf.shape[2] - 1) / 2),
                              bias=False)
    with torch.no_grad():
        conv_op.weight = torch.nn.Parameter(psf)
    # mse = torch.nn.MSELoss()
    mse_ = 0
    conv_img = conv_op(deblurred_image)
    for i in range(blurry_image.shape[2]):
        mse_ += detector_weights[i]*torch.sum(torch.square(blurry_image[:, :, i, :, :] - conv_img))
    return mse_ / (a * b * h * w)


class SpitfireJoinDeconv(SAiryscanReconstruction):
    """Gray scaled image deconvolution with the Spitfire algorithm

    :param psf:  Point spread function
    :param weight: Model weight between hessian and sparsity. Value is in  ]0, 1[
    :param reg: Regularization weight. Value is in [0, 1]
    :param detector_weights: Method to compute the detectors weights
    """
    def __init__(self,
                 psf: torch.Tensor,
                 weight: float = 0.6,
                 reg: float = 0.5,
                 detector_weights: str = 'mean'):
        super().__init__()
        self.num_args = 1
        self.psf = psf
        self.weight = weight
        self.reg = reg
        self.detector_weights = detector_weights
        self.niter_ = 0
        self.max_iter_ = 2000
        self.gradient_step_ = 0.01
        self.loss_ = None

    def __call__(self, image: torch.Tensor) -> torch.Tensor:
        """Reconstruct the spitfire join deconvolution

        :param image: Raw airyscan image. [H, Z, Y, X] for 3D image, [H, Y, X] for 2D images,
        :return: The reconstructed image. [Z, Y, X] for 3D, [Y, X] for 2D
        """
        if image.ndim == 3:
            return self.run_2d(image)
        if image.ndim == 4:
            raise NotImplementedError('Spitfire 3D deconvolution is not yet implemented')
        raise NotImplementedError('Spitfire reconstruction: image dimension not supported')

    def run_2d(self, image) -> torch.Tensor:
        """Do the 2D reconstruction

        :param image: Raw airyscan image. [H, Z, Y, X] for 3D image, [H, Y, X] for 2D images,
        :return: The reconstructed image. [Z, Y, X] for 3D, [Y, X] for 2D
        """
        self.progress(0)

        # detectors weights
        detector_weights_filter = SAiryscanWeights(self.detector_weights)
        detector_weights_ = detector_weights_filter()

        # normalise image
        image = (image - torch.min(image)) / (torch.max(image) - torch.min(image))

        # deconv ref image
        deconv_image = torch.zeros(image.shape[1], image.shape[2])
        for d in range(32):
            deconv_image = deconv_image + detector_weights_[d]*image[d, ...]
        # ism_image = torch.mean(image, axis=0)
        deconv_image = deconv_image.view(1, 1, deconv_image.shape[0], deconv_image.shape[1])
        deconv_image.requires_grad = True

        image = image.view(1, 1, image.shape[0], image.shape[1], image.shape[2])
        self.psf = self.psf.view(1, 1, self.psf.shape[0], self.psf.shape[1])

        optimizer = torch.optim.Adam([deconv_image], lr=self.gradient_step_)
        scheduler = torch.optim.lr_scheduler.StepLR(optimizer, step_size=100, gamma=0.5)
        previous_loss = 9e12
        count_eq = 0
        self.niter_ = 0

        loss = None
        for i in range(self.max_iter_):
            self.progress(int(100*i/self.max_iter_))
            self.niter_ += 1
            optimizer.zero_grad()
            loss = self.reg * dataterm_deconv(image, deconv_image, self.psf, detector_weights_) + \
                (1-self.reg) * hv_loss(deconv_image, self.weight)
            print('iter:', self.niter_, ' loss:', loss.item())
            if abs(loss - previous_loss) < 1e-7:
                count_eq += 1
            else:
                previous_loss = loss
                count_eq = 0
            if count_eq > 5:
                break
            loss.backward()
            optimizer.step()
            scheduler.step()
        self.loss_ = loss
        self.progress(100)
        return self._crop(deconv_image.view(image.shape[3], image.shape[4]))


metadata = {
    'name': 'SpitfireJoinDeconv',
    'label': 'Spitfire Join Deconv',
    'class': SpitfireJoinDeconv,
    'parameters': {
        'psf': {
            'type': torch.Tensor,
            'label': 'psf',
            'help': 'Point Spread Function',
            'default': None
        },
        'detector_weights': {
            'type': 'select',
            'label': 'Detectors weights',
            'help': 'Model defining weight for each detectors',
            'values': ['mean', 'ring', 'ring_inv', 'd2c', 'id2c', 'exp_d2c', 'exp_d2c_inv'],
            'default': 'mean'
        },
        'weight': {
            'type': float,
            'label': 'Weight',
            'help': 'Model weight between hessian and sparsity. Value is in  ]0, 1[',
            'default': 0.6,
            'range': (0, 1)
        },
        'reg': {
            'type': float,
            'label': 'Regularization',
            'help': 'Regularization weight. Value is in [0, 1]',
            'default': 0.995,
            'range': (0, 1)
        }
    }
}



"""This module implements the generation of Gaussian PSFs"""
import math
import numpy as np
import torch


class PSFGaussian:
    """Generate a Gaussian PSF

    :param sigma: Radius of the Gaussian in each dimension
    :param shape: Size of the PSF array in each dimension
    """
    def __init__(self, sigma: tuple[int, ...], shape: tuple[int, ...]):
        self.sigma = sigma
        self.shape = shape
        self.psf_ = None

    def __call__(self) -> torch.Tensor:
        """Calculate the PSF image"""
        if len(self.shape) == 2:
            self.psf_ = np.zeros(self.shape)
            x0 = math.floor(self.shape[0] / 2)
            y0 = math.floor(self.shape[1] / 2)
            sigma_x2 = 0.5 / (self.sigma[0] * self.sigma[0])
            sigma_y2 = 0.5 / (self.sigma[1] * self.sigma[1])
            for x in range(self.shape[0]):
                for y in range(self.shape[1]):
                    self.psf_[x, y] = math.exp(- pow(x-x0, 2) * sigma_x2
                                               - pow(y-y0, 2) * sigma_y2)
        elif len(self.shape) == 3:
            x0 = self.shape[0] / 2
            y0 = self.shape[1] / 2
            z0 = self.shape[2] / 2
            sigma_x2 = 0.5 / self.sigma[0] * self.sigma[0]
            sigma_y2 = 0.5 / self.sigma[1] * self.sigma[1]
            sigma_z2 = 0.5 / self.sigma[2] * self.sigma[2]
            for x in range(self.shape[0]):
                for y in range(self.shape[1]):
                    for z in range(self.shape[2]):
                        self.psf_[z, x, y] = math.exp(- pow(x-x0, 2) * sigma_x2
                                                      - pow(y-y0, 2) * sigma_y2
                                                      - pow(z-z0, 2) * sigma_z2)
        else:
            raise ValueError('PSFGaussian: can generate only 2D or 3D PSFs')
        return torch.tensor(self.psf_/np.sum(self.psf_)).float()



"""This module implements the IFED reconstruction method"""
import torch
from .interface import SAiryscanReconstruction
from ._sure import SureMap


class IFED(SAiryscanReconstruction):
    """Airyscan reconstruction using the IFED method

    :param inner_ring_index: Index of the last detector of the inner ring [7, 19]
    :param epsilon: Coefficient (or weight) for combining inner and outer ring images
    """
    def __init__(self, inner_ring_index: int = 7, epsilon: str | float = 0.3):
        super().__init__()

        print('ifed constructor inner index=', inner_ring_index)
        print('ifed constructor epsilon=', epsilon)
        self.num_args = 1
        self.inner_ring_index = inner_ring_index
        self.epsilon = epsilon
        if inner_ring_index not in [7, 19]:
            raise ValueError('Inner ring index must be in (7, 19)')

    def __call__(self, image: torch.Tensor) -> torch.Tensor:
        """Do the reconstruction

        :param image: Raw detector stack to reconstruct [H (Z) Y X]
        :return: high resolution image [(Z) Y X]
        """
        self.progress(0)
        self.notify('IFED: sum inner and outer detectors')
        a = torch.sum(image[0:self.inner_ring_index, ...], axis=0)
        b = torch.sum(image[self.inner_ring_index + 1:32, ...], axis=0)

        self.progress(33)
        print('ifed epsilon=', self.epsilon)
        if self.epsilon == 'map':
            self.notify('IFED: compute epsilon map')
            map_ = SureMap(smooth=True)
            epsilon = map_(image[0, ...], a, b)
            print('epsilon map=', epsilon.shape)
        elif self.epsilon == 'mode':
            self.notify('IFED: estimate epsilon using sure')
            map_ = SureMap(smooth=False)
            epsilon = map_.mode(image[0, ...], a, b)
            print('mode epsilon=', epsilon)
        else:
            self.notify('IFED: use manual epsilon')
            epsilon = torch.tensor(self.epsilon)
        self.progress(75)
        self.notify('IFED: do sum and relu')
        out = a - epsilon * b
        self.progress(100)
        return self._crop(torch.nn.functional.relu(out, inplace=True))


metadata = {
    'name': 'IFED',
    'label': 'IFED',
    'class': IFED,
    'parameters': {
        'inner_ring_index': {
            'type': int,
            'label': 'Inner index',
            'help': 'Index of the inner ring last detector (7, 19)',
            'default': 7
        },
        'epsilon': {
            'type': str,
            'label': 'epsilon',
            'help': 'Weighting parameter',
            'default': 0.3
        }
    }
}



"""This module implements the Spitfire workflow for denoising or deconvolution"""
import numpy as np

from scipy.signal import convolve2d
import torch
import torch.nn.functional as F

from .interface import SAiryscanReconstruction


def hv_loss(img: torch.Tensor, weighting: float) -> torch.Tensor:
    """Sparse Hessian regularization term

    :param img: Tensor of shape BCYX containing the estimated image,
    :param weighting: Sparse weighting parameter in [0, 1]. 0 sparse, and 1 not sparse,
    :return: The regularization value
    """
    a, b, h, w = img.size()
    dxx2 = torch.square(-img[:, :, 2:, 1:-1] + 2 * img[:, :, 1:-1, 1:-1] - img[:, :, :-2, 1:-1])
    dyy2 = torch.square(-img[:, :, 1:-1, 2:] + 2 * img[:, :, 1:-1, 1:-1] - img[:, :, 1:-1, :-2])
    dxy2 = torch.square(img[:, :, 2:, 2:] - img[:, :, 2:, 1:-1] - img[:, :, 1:-1, 2:] +
                        img[:, :, 1:-1, 1:-1])
    hv = torch.sqrt(weighting * weighting * (dxx2 + dyy2 + 2 * dxy2) +
                    (1 - weighting) * (1 - weighting) * torch.square(img[:, :, 1:-1, 1:-1])).sum()
    return hv / (a * b * h * w)


def hessian(img: torch.Tensor) -> torch.Tensor:
    """Apply Hessian filter on a tensor

    :param img: Input image tensor,
    :return: Tensor of filtered image
    """
    dxx2 = torch.square(-img[:, :, 2:, 1:-1] + 2 * img[:, :, 1:-1, 1:-1] - img[:, :, :-2, 1:-1])
    dyy2 = torch.square(-img[:, :, 1:-1, 2:] + 2 * img[:, :, 1:-1, 1:-1] - img[:, :, 1:-1, :-2])
    dxy2 = torch.square(img[:, :, 2:, 2:] - img[:, :, 2:, 1:-1] - img[:, :, 1:-1, 2:] +
                        img[:, :, 1:-1, 1:-1])
    return F.pad(dxx2 + dyy2 + 2 * dxy2, (1, 1, 1, 1), "constant", 0)


def dataterm_flow(blurry_image: torch.Tensor,
                  deblurred_image: torch.Tensor,
                  detector_weights: torch.Tensor
                  ) -> torch.Tensor:
    """Denoising L2 data-term

    Compute the L2 error between the original image and the reconstructed image flow

    :param blurry_image: Tensor of shape BCHYX containing the original blurry image
    :param deblurred_image: Tensor of shape BCYX containing the estimated deblurred image
    :param detector_weights: Weight applied to each detector
    """
    mse = torch.nn.MSELoss()
    mse_ = 0
    for i in range(blurry_image.shape[2]):
        mse_ += mse(blurry_image[:, :, i, :, :] -
                    detector_weights[i]*hessian(blurry_image[:, :, i, :, :]), deblurred_image)
    return mse_


def estimate_noise(data: torch.Tensor) -> float:
    """ Estimate the RMS noise of an image

    from http://stackoverflow.com/questions/2440504/
                noise-estimation-noise-measurement-in-image

    Reference: J. Immerkaer, “Fast Noise Variance Estimation”,
    Computer Vision and Image Understanding,
    Vol. 64, No. 2, pp. 300-302, Sep. 1996 [PDF]

    :param data: Image to estimate noise
    :return: The estimated variance of the noise
    """
    data = data.detach().numpy()[0, 0, ...]
    print('data.shape=', data.shape)
    height, width = data.shape
    data = np.nan_to_num(data)
    mat = [[1, -2, 1],
           [-2, 4, -2],
           [1, -2, 1]]

    sigma = np.sum(np.sum(np.abs(convolve2d(data, mat))))
    sigma = sigma * np.sqrt(0.5 * np.pi) / (6 * (width - 2) * (height - 2))

    return sigma*sigma


def calculate_weights(image: torch.Tensor) -> torch.Tensor:
    """Calculate the hessian weights of each detector

    :param image: Detectors raw images
    :return: vector containing one weight per detector
    """
    weights = torch.ones((32,))
    for d in range(32):
        weights[d] = estimate_noise(image[:, :, d, :, :])
    return weights


class SSpitfireFlow(SAiryscanReconstruction):
    """Workflow for spitfire denoising

    :param weight: Spitfire weight in ]0, 1[,
    :param reg: Spitfire regularization in ]0, 1[
    """
    def __init__(self, weight: float = 0.6, reg: float = 0.5):
        super().__init__()
        self.num_args = 1
        self.weight = weight
        self.reg = reg
        self.niter_ = 0
        self.max_iter_ = 2000
        self.gradient_step_ = 0.01
        self.loss_ = None

    def __call__(self, image: torch.Tensor) -> torch.Tensor:
        """Reconstruct the opticalflow denoised ISM with spitfire regularisation

        :param image: Raw airyscan image. [H, Z, Y, X] for 3D image, [H, Y, X] for 2D images
        :return: The reconstructed image. [Z, Y, X] for 3D, [Y, X] for 2D
        """
        if image.ndim == 3:
            return self.run_2d(image)
        if image.ndim == 4:
            raise NotImplementedError('Spitfire 3D deconvolution is not yet implemented')
        raise NotImplementedError('Spitfire reconstruction: image dimension not supported')

    def run_2d(self, image: torch.Tensor) -> torch.Tensor:
        """Implements the 2D case

        :param image: Raw airyscan image. [H, Z, Y, X] for 3D image, [H, Y, X] for 2D images
        :return: The reconstructed image. [Z, Y, X] for 3D, [Y, X] for 2D
        """
        image = image.view(1, 1, image.shape[0], image.shape[1], image.shape[2])
        deconv_image = image[:, :, 0, :, :].detach().clone()
        deconv_image.requires_grad = True
        optimizer = torch.optim.Adam([deconv_image], lr=self.gradient_step_)
        scheduler = torch.optim.lr_scheduler.StepLR(optimizer, step_size=100, gamma=0.5)
        previous_loss = 9e12
        count_eq = 0
        self.niter_ = 0
        detector_weights_ = calculate_weights(image)
        loss = None
        for _ in range(self.max_iter_):
            self.niter_ += 1
            optimizer.zero_grad()
            loss = self.reg * dataterm_flow(image, deconv_image, detector_weights_) + \
                (1-self.reg) * hv_loss(deconv_image, self.weight)
            print('iter:', self.niter_, ' loss:', loss)
            if abs(loss - previous_loss) < 1e-7:
                count_eq += 1
            else:
                previous_loss = loss
                count_eq = 0
            if count_eq > 5:
                break
            loss.backward()
            optimizer.step()
            scheduler.step()
        self.loss_ = loss
        return self._crop(deconv_image.view(image.shape[3], image.shape[4]))


metadata = {
    'name': 'SSpitfireFlow',
    'label': 'Spitfire Flow',
    'class': SSpitfireFlow,
    'parameters': {
        'weight': {
            'type': float,
            'label': 'Weight',
            'help': 'Model weight between hessian and sparsity. Value is in  ]0, 1[',
            'default': 0.6,
            'range': (0, 1)
        },
        'reg': {
            'type': float,
            'label': 'Regularization',
            'help': 'Regularization weight. Value is in [0, 1]',
            'default': 0.5,
            'range': (0, 1)
        }
    }
}



"""Airyscan IO

Set of IO methods to read Airyscan raw files
"""
from pathlib import Path
import numpy as np
import torch
from aicspylibczi import CziFile
from .settings import Settings


class SAiryscanReader:
    """Read a CZI Zeiss raw data

    :param filename: Path to the .czi file

    """
    def __init__(self, filename: str):
        super().__init__()
        if filename.endswith('.czi') or filename.endswith('.CZI'):
            self.filename = filename
        else:
            raise IOError('AiryscanReader can only read CZI files')

        pth = Path(filename)
        self.czi = CziFile(pth)
        self.dimensions = self.czi.get_dims_shape()[0]
        print('pixel type=', self.czi.pixel_type)
        print('dimensions=', self.dimensions)

    def frames(self) -> int:
        """Get the number of frames in the CZI file"""
        return self.dimensions['T'][1]

    def channels(self) -> int:
        """Get the number of channels in the CZI file"""
        return self.dimensions['C'][1]

    def depth(self) -> int:
        """Get the number of slices in the CZI file"""
        return self.dimensions['Z'][1]

    def width(self) -> int:
        """Get the number of columns in the CZI file"""
        return self.dimensions['X'][1]

    def height(self) -> int:
        """Get the number of rows in the CZI file"""
        return self.dimensions['Y'][1]

    def data(self, frame_idx: int = 0, channel_idx: int = 0):
        """Extract one frame from the CZI data

        :param frame_idx: Frame index
        :param channel_idx: Channel index
        :return: a torch array [H, Z, Y, X] for 3D slice, [H, Y, X] for 2D slices. Where H is the
                 detector dimension
        """
        img, _ = self.czi.read_image(T=frame_idx, C=channel_idx)
        return torch.from_numpy(np.float32(np.squeeze(img))).to(Settings.instance().device)



"""This module implement image denoising by applying a Gaussian filter"""
import torch
from torchvision import transforms

from .interface import SAiryscanEnhancing


class SAiryscanGaussian(SAiryscanEnhancing):
    """Apply a gaussian filter

    :param sigma: Gaussian sigma in each dimension
    :param kernel_size: Size of the patch used for the Gaussian support
    """
    def __init__(self, sigma: float | tuple[float, ...] = 0.5, kernel_size: int =7):
        super().__init__()
        self.sigma = sigma
        self.kernel_size = kernel_size

    def __call__(self, image: torch.Tensor) -> torch.Tensor:
        """Run the filtering

        :param image: Image to filter
        :return: The filtered image
        """
        out_bc = image.view(1, 1, image.shape[0], image.shape[1])
        out_b = transforms.functional.gaussian_blur(out_bc,
                                                    kernel_size=(self.kernel_size,
                                                                 self.kernel_size),
                                                    sigma=self.sigma)
        return out_b.view(image.shape)


metadata = {
    'name': 'SAiryscanGaussian',
    'label': 'Gaussian Filter',
    'class': SAiryscanGaussian,
    'parameters': {
        'sigma': {
            'type': float,
            'label': 'Sigma',
            'help': 'Gaussian sigma parameter',
            'default': 0.5,
            'range': (0, 512)
        },
        'kernel_size': {
            'type': int,
            'label': 'kernel size',
            'help': 'Size of the kernel in the x and y directions',
            'default': 7,
            'range': (3, 512)
        }
    }
}



"""Interface for Airyscan reconstruction process"""
import torch
from sairyscan.core import SObservable


class SAiryscanReconstruction(SObservable):
    """Interface for Airyscan high resolution image reconstruction method"""
    def __call__(self, image: torch.Tensor, *args, **kwargs) -> torch.Tensor:
        """Do the reconstruction

        :param image: Raw detector stack to reconstruct [H (Z) Y X]
        :return: high resolution image [(Z) Y X]
        """
        raise NotImplementedError('SairyscanReconstruction is an interface. Please implement the'
                                  ' __call__ method')

    @staticmethod
    def _crop(image: torch.Tensor) -> torch.Tensor:
        """Crop the image by 5 pixels to avoid side artifacts

        :param image: Image to crop,
        :return: Cropped image
        """
        if image.ndim == 2:
            return image[5:-5, 5:-5]
        return image[5:-5, 5:-5, :]



"""Interface for Airyscan enhancing filter"""
import torch

from sairyscan.core import SObservable


class SAiryscanEnhancing(SObservable):
    """Interface for Airyscan enhancing filter applied after the reconstruction"""
    def __call__(self, image: torch.Tensor) -> torch.Tensor:
        """Do the enhancing

        :param image: Reconstructed airyscan data for a single channel time point [(Z) Y X]
        :return: Filtered image
        """
        raise NotImplementedError('SairyscanEnhancing is an interface. Please implement the'
                                  ' __call__ method')



"""Implementation of detectors registration using the Fourier Phase loss"""
import torch
import torch.nn.functional as F
from torchvision import transforms
from skimage.registration import phase_cross_correlation
from skimage.transform import rescale

from .interface import SAiryscanRegistration


class SRegisterFourierPhase(SAiryscanRegistration):
    """Register the detectors stack by translating each image to the detector position is array

    :param weight: Weight applied on the detector position translation

    """
    def __init__(self, weight: float = 1.0):
        super().__init__()
        self.weight = weight

    def __call__(self, image: torch.Tensor) -> torch.Tensor:
        """Do the registration

        :param image: Single channel/frame stack to register,
        :return: The registered stack
        """
        # image_np = image.detach().numpy()
        image_out = torch.zeros(image.shape, dtype=torch.float32)
        image_out[0, ...] = image[0, ...].clone()
        ratio = (image.shape[1]-5)/image.shape[1]
        for i in range(1, 32):
            shift, _, _ = phase_cross_correlation(
                                            rescale(image[0, ...].detach().numpy(), 2),
                                            rescale(image[i, ...].detach().numpy(), 2),
                                            normalization=None,
                                            overlap_ratio=ratio
                                            )
            print(f'shift {i}= [{shift[1] / 2 }, {shift[0] / 2 }]')
            image_out[i, ...] = self._translate_detector(image[i, ...], (shift[1]/2, shift[0]/2))
        return image_out

    @staticmethod
    def hessian(img: torch.Tensor) -> torch.Tensor:
        """Sparse Hessian regularization term

        :param img: Tensor of shape BCYX containing the estimated image
        :return: THe hessian term
        """
        dxx2 = torch.square(-img[2:, 1:-1] + 2 * img[1:-1, 1:-1] - img[:-2, 1:-1])
        dyy2 = torch.square(-img[1:-1, 2:] + 2 * img[1:-1, 1:-1] - img[1:-1, :-2])
        dxy2 = torch.square(img[2:, 2:] - img[2:, 1:-1] - img[1:-1, 2:] +
                            img[1:-1, 1:-1])
        return dxx2 + dyy2 + 2 * dxy2

    @staticmethod
    def gradient(img: torch.Tensor) -> torch.Tensor:
        """Compute the gradient of the image

        :param img: Image to process,
        :return: THe gradient image of img
        """
        tv_h = torch.abs(img[1:, :] - img[:-1, :])
        tv_w = torch.abs(img[:, 1:] - img[:, :-1])
        return F.pad(tv_h, (0, 0, 1, 0), "constant", 0) + F.pad(tv_w, (1, 0, 0, 0), "constant", 0)

    @staticmethod
    def _translate_detector(img: torch.Tensor, translate: tuple[float, float]) -> torch.Tensor:
        """Translate one detector

        :param img: Tensor for one single detector
        :param translate: (x, y, z) or (x, y) translation vector
        :returns: the translated tensor
        """
        img_bc = img.view((1, 1, img.shape[0], img.shape[1]))
        img_out = transforms.functional.affine(img_bc, angle=0, translate=translate, scale=1,
                    shear=[0, 0],
                    interpolation=transforms.functional.InterpolationMode.BILINEAR)
        return img_out.view(img.shape)


metadata = {
    'name': 'SRegisterFourierPhase',
    'class': SRegisterFourierPhase,
    'parameters': {
    }
}



"""This module implements factory to instantiate models"""
import numpy as np
import torch

from ..core import SObservable


class SAiryscanFactoryError(Exception):
    """Raised when an error happen when a module is built in the factory"""


class SAiryscanModuleFactory:
    """Factory for SAiryscan modules"""
    def __init__(self):
        self._data = {}

    def register(self, key: str, metadata: dict[str, any]):
        """Register a new builder to the factory

        :param key: Name of the module to register,
        :param metadata: Dictionary containing the filter metadata
        """
        self._data[key] = metadata

    def get_parameters(self, key: str) -> dict[str, any]:
        """Get the parameters of a filter

        :param key: Name of the filter
        :return: The dictionary of parameters
        """
        return self._data[key]['parameters']

    def get_keys(self) -> list[str]:
        """Get the names of all the registered modules

        :return: The list of all the registered modules names
        """
        return list(self._data.keys())

    def get(self, key: str, **kwargs: dict[str, any]) -> SObservable:
        """Get the instance of the SAiryscan module

        :param key: Name of the module to load
        :param kwargs: Dictionary of CLI args for models parameters (ex: number of channels)
        :return: The module instance
        """
        metadata = self._data.get(key)
        if not metadata:
            raise ValueError(key)
        builder = SAiryscanModuleBuilder()
        return builder.get_instance(metadata, kwargs)


class SAiryscanModuleBuilder:
    """Interface for a SAiryscan module builder

    The builder is used by the factory to instantiate a module

    """
    def __init__(self):
        self._instance = None

    def get_instance(self, metadata: dict[str, any], args: dict[str, any]):
        """Get the instance of the module

        :param metadata: Metadata of the module
        :param args: Arguments to instantiate the module
        :return: instance of the module

        """
        # check the args
        instance_args = {}
        for key, value in metadata['parameters'].items():
            val = self._get_arg(value, key, args)
            instance_args[key] = val
        return metadata['class'](**instance_args)

    def _get_arg(self,
                 param_metadata: dict[str, any],
                 key: str,
                 args: dict[str, any]
                 ) -> int | float | str | bool | torch.Tensor:
        """Check and convert an argument value from parameters dictionary

        :param param_metadata: Information on the parameter
        :param key: Name of the parameter
        :param args: Value of the parameter
        :return: The value of the parameter in the proper type and format
        """
        type_ = param_metadata['type']
        print('param metadata:', param_metadata)
        range_ = None
        if 'range' in param_metadata:
            range_ = param_metadata['range']

        if type_ is float:
            return self.get_arg_float(args, key, param_metadata['default'],
                                      range_)
        if type_ is int:
            return self.get_arg_int(args, key, param_metadata['default'],
                                    range_)
        if type_ is bool:
            return self.get_arg_bool(args, key, param_metadata['default'],
                                     range_)
        if type_ is str:
            return self.get_arg_str(args, key, param_metadata['default'])
        if type_ is torch.Tensor:
            return self.get_arg_array(args, key, param_metadata['default'])
        if type_ == 'select':
            return self.get_arg_select(args, key, param_metadata['values'])
        raise ValueError(f"parameter type '{type_}' not recognized")

    @staticmethod
    def _error_message(key: str, value_type: str, value_range: tuple | None) -> str:
        """Build the error message to raise if an input parameter is not correct

        :param key: Input parameter key
        :param value_type: String naming the input type (int, float...)
        :param value_range: Min and max values of the parameter
        :return: The error message
        """
        range_message = ''
        if value_range and len(value_range) == 2:
            range_message = f' in range [{str(value_range[0]), str(value_range[1])}]'

        message = f'Parameter {key} must be of type `{value_type}` {range_message}'
        return message

    def get_arg_int(self,
                    args: dict[str, any],
                    key: str,
                    default_value: int,
                    value_range: tuple[int, ...] = None
                    ) -> int:
        """Get the integer value of a parameter from the args list

        The default value of the parameter is returned if the
        key is not in args.

        :param args: Dictionary of the input args,
        :param key: Name of the parameters,
        :param default_value: Default value of the parameter,
        :param value_range: Min and max value of the parameter,
        :return: The parameter value
        """
        value = default_value
        if isinstance(args, dict) and key in args:
            # cast
            try:
                value = int(args[key])
            except ValueError as err:
                raise SAiryscanFactoryError(self._error_message(key, 'int', value_range)) from err
        # test range
        if value_range and len(value_range) == 2:
            if value > value_range[1] or value < value_range[0]:
                raise SAiryscanFactoryError(self._error_message(key, 'int', value_range))
        return value

    def get_arg_float(self,
                      args: dict[str, any],
                      key: str,
                      default_value: float,
                      value_range: tuple[float, ...] = None
                      ) -> float:
        """Get the float value of a parameter from the args list

        The default value of the parameter is returned if the
        key is not in args.

        :param args: Dictionary of the input args,
        :param key: Name of the parameters,
        :param default_value: Default value of the parameter,
        :param value_range: Min and max value of the parameter,
        :return: The parameter value
        """
        value = default_value
        if isinstance(args, dict) and key in args:
            # cast
            try:
                value = float(args[key])
            except ValueError as err:
                raise SAiryscanFactoryError(self._error_message(key, 'float', value_range)) from err
        # test range
        if value_range and len(value_range) == 2:
            if value > value_range[1] or value < value_range[0]:
                raise SAiryscanFactoryError(self._error_message(key, 'float', value_range))
        return value

    def get_arg_str(self,
                    args: dict[str, any],
                    key: str,
                    default_value: str,
                    value_range: tuple[str, ...] = None):
        """Get the string value of a parameter from the args list

        The default value of the parameter is returned if the
        key is not in args.

        :param args: Dictionary of the input args,
        :param key: Name of the parameters,
        :param default_value: Default value of the parameter,
        :param value_range: Min and max value of the parameter,
        :return The parameter value
        """
        value = default_value
        if isinstance(args, dict) and key in args:
            # cast
            try:
                value = str(args[key])
            except ValueError as err:
                raise SAiryscanFactoryError(self._error_message(key, 'str', value_range)) from err
        # test range
        if value_range and len(value_range) == 2:
            if value > value_range[1] or value < value_range[0]:
                raise SAiryscanFactoryError(self._error_message(key, 'str', value_range))
        return value

    def get_arg_bool(self,
                     args: dict[str, any],
                     key: str,
                     default_value: bool,
                     value_range: tuple[bool, ...] = None
                     ) -> bool:
        """Get the boolean value of a parameter from the args list

        The default value of the parameter is returned if the
        key is not in args.

        :param args: Dictionary of the input args,
        :param key: Name of the parameters,
        :param default_value: Default value of the parameter,
        :param value_range: Min and max value of the parameter,
        :return: The parameter value
        """
        value = default_value
        # cast
        if isinstance(args, dict) and key in args:
            if isinstance(args[key], str):
                if args[key] == 'True':
                    value = True
                else:
                    value = False
            elif isinstance(args[key], bool):
                value = args[key]
            else:
                raise SAiryscanFactoryError(self._error_message(key, 'bool', value_range))
        # test range
        if value_range and len(value_range) == 2:
            if value > value_range[1] or value < value_range[0]:
                raise SAiryscanFactoryError(self._error_message(key, 'bool', value_range))
        return value

    def get_arg_array(self,
                      args: dict[str, any],
                      key: str,
                      default_value: torch.Tensor
                      ) -> torch.Tensor:
        """Get the value of a parameter from the args list

        The default value of the parameter is returned if the
        key is not in args.

        :param args: Dictionary of the input args,
        :param key: Name of the parameters,
        :param default_value: Default value of the parameter,
        :return: The parameter value
        """
        value = default_value
        if isinstance(args, dict) and key in args:
            # rint('psf type=', type(args[key]))
            if isinstance(args[key], torch.Tensor):
                value = args[key]
            elif isinstance(args[key], np.ndarray):
                value = torch.Tensor(args[key])
            else:
                raise SAiryscanFactoryError(self._error_message(key, 'array', None))
        return value

    def get_arg_select(self,
                       args: dict[str, any],
                       key: str,
                       values: list[str]
                       ) -> str:
        """Get the select value of a parameter from the args list

        The default value of the parameter is returned if the
        key is not in args.

        :param args: Dictionary of the input args,
        :param key: Name of the parameters,
        :param values: Default value of the parameter,
        :return: The parameter value
        """
        if isinstance(args, dict) and key in args:
            value = str(args[key])
            for x in values:
                if str(x) == value:
                    return x
        raise SAiryscanFactoryError(self._error_message(key, 'select', None))



"""This module implements image enhancing methods (denoising and deconvolution)"""
from .interface import SAiryscanEnhancing
from .gaussian import SAiryscanGaussian
from .wiener import SAiryscanWiener
from .richardson_lucy import SAiryscanRichardsonLucy
from ._psfs import PSFGaussian
from .spitfire_deconv import SpitfireDeconv

from .gaussian import metadata as gaussian_metadata
from .richardson_lucy import metadata as richardson_lucy_metadata
from .wiener import metadata as wiener_metadata
from .spitfire_deconv import metadata as spitfiredeconv_metadata
from .spitfire_denoise import metadata as spitfiredenoise_metadata


metadata = [gaussian_metadata,
            spitfiredenoise_metadata,
            richardson_lucy_metadata,
            wiener_metadata,
            spitfiredeconv_metadata
            ]


__all__ = [
    'PSFGaussian',
    'SAiryscanEnhancing',
    'SAiryscanGaussian',
    'SAiryscanWiener',
    'SAiryscanRichardsonLucy',
    'SpitfireDeconv',
    'metadata'
]



"""This module implements the read/write of user settings"""
import torch


class SettingsContainer:
    """Container for the SAiryscan library settings"""
    def __init__(self):
        self.__device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")

    @property
    def device(self) -> str:
        """Get the device ID

        :return: The device ID
        """
        return self.__device


class Settings:
    """Singleton to access the Settings container
        
    :raises: RuntimeError if multiple instantiation of the Settings container is tried
    """
    __instance = None

    def __init__(self):
        """ Virtually private constructor. """
        if Settings.__instance is not None:
            raise RuntimeError("Settings container can be initialized only once!")
        Settings.__instance = SettingsContainer()

    @staticmethod
    def instance():
        """ Static access method to the Config. """
        if Settings.__instance is None:
            Settings.__instance = SettingsContainer()
        return Settings.__instance



"""This module implement a Spitfire deconvolution"""
import torch
from .interface import SAiryscanEnhancing


def hv_loss(img: torch.Tensor, weighting: float) -> torch.Tensor:
    """Sparse Hessian regularization term

    :param img: Tensor of shape BCYX containing the estimated image
    :param weighting: Sparse weighting parameter in [0, 1]. 0 sparse, and 1 not sparse
    :return: The loss value
    """
    a, b, h, w = img.size()
    dxx2 = torch.square(-img[:, :, 2:, 1:-1] + 2 * img[:, :, 1:-1, 1:-1] - img[:, :, :-2, 1:-1])
    dyy2 = torch.square(-img[:, :, 1:-1, 2:] + 2 * img[:, :, 1:-1, 1:-1] - img[:, :, 1:-1, :-2])
    dxy2 = torch.square(img[:, :, 2:, 2:] - img[:, :, 2:, 1:-1] - img[:, :, 1:-1, 2:] +
                        img[:, :, 1:-1, 1:-1])
    hv = torch.sqrt(weighting * weighting * (dxx2 + dyy2 + 2 * dxy2) +
                    (1 - weighting) * (1 - weighting) * torch.square(img[:, :, 1:-1, 1:-1])).sum()
    return hv / (a * b * h * w)


def dataterm_deconv(blurry_image: torch.Tensor, deblurred_image: torch.Tensor, psf: torch.Tensor):
    """Deconvolution L2 data-term

    Compute the L2 error between the original image and the convoluted reconstructed image.

    :param blurry_image: Tensor of shape BCYX containing the original blurry image,
    :param deblurred_image: Tensor of shape BCYX containing the estimated deblurred image,
    :param psf: Tensor containing the point spread function,
    :return: The data term value
    """
    conv_op = torch.nn.Conv2d(1, 1, kernel_size=psf.shape[2],
                              stride=1,
                              padding=int((psf.shape[2] - 1) / 2),
                              bias=False)
    with torch.no_grad():
        conv_op.weight = torch.nn.Parameter(psf)
    mse = torch.nn.MSELoss()
    return mse(blurry_image, conv_op(deblurred_image))


class SpitfireDeconv(SAiryscanEnhancing):
    """Gray scaled image deconvolution with the Spitfire algorithm

    :param psf: Point spread function,
    :param weight: Model weight between hessian and sparsity. Value is in  ]0, 1[,
    :param reg: Regularization weight. Value is in [0, 1]
    """
    def __init__(self, psf: torch.Tensor, weight: float = 0.6, reg: float = 0.995):
        super().__init__()
        self.psf = psf
        self.weight = weight
        self.reg = reg
        self.niter_ = 0
        self.max_iter_ = 2000
        self.gradient_step_ = 0.01
        self.loss_ = None

    def __call__(self, image: torch.Tensor) -> torch.Tensor:
        """Do the deconvolution

        :param image: Image to deblur
        :return: The deblurred image
        """
        if image.ndim == 2:
            return self.run_2d(image)
        raise NotImplementedError('Spitfire 3D deconvolution is not yet implemented')

    def run_2d(self,  image: torch.Tensor) -> torch.Tensor:
        """Do the deconvolution for the 2D case

        :param image: Image to deblur
        :return: The deblurred image
        """
        self.progress(0)
        image = (image-torch.min(image))/(torch.max(image)-torch.min(image))
        # pad image
        padding = 13
        pad_fn = torch.nn.ReflectionPad2d(padding)
        image_pad = pad_fn(image.detach().clone().view(1, 1, image.shape[0], image.shape[1]))

        self.psf = self.psf.view(1, 1, self.psf.shape[0], self.psf.shape[1])
        deconv_image = image_pad.detach().clone()
        deconv_image.requires_grad = True
        optimizer = torch.optim.Adam([deconv_image], lr=self.gradient_step_)
        scheduler = torch.optim.lr_scheduler.StepLR(optimizer, step_size=100, gamma=0.5)
        previous_loss = 9e12
        count_eq = 0
        self.niter_ = 0
        loss = None
        for i in range(self.max_iter_):
            self.progress(int(100*i/self.max_iter_))
            self.niter_ += 1
            optimizer.zero_grad()
            loss = self.reg * dataterm_deconv(image_pad, deconv_image, self.psf) + \
                (1-self.reg) * hv_loss(deconv_image, self.weight)
            print('iter:', self.niter_, ' loss:', loss.item())
            if abs(loss - previous_loss) < 1e-7:
                count_eq += 1
            else:
                previous_loss = loss
                count_eq = 0
            if count_eq > 5:
                break
            loss.backward()
            optimizer.step()
            scheduler.step()
        self.loss_ = loss
        self.progress(100)
        return deconv_image.view(image_pad.shape[2],
                                 image_pad.shape[3])[padding:-padding, padding:-padding]


metadata = {
    'name': 'SpitfireDeconv',
    'label': 'Spitfire Deconv',
    'class': SpitfireDeconv,
    'parameters': {
        'psf': {
            'type': torch.Tensor,
            'label': 'psf',
            'help': 'Point Spread Function',
            'default': None
        },
        'weight': {
            'type': float,
            'label': 'weight',
            'help': 'Model weight between hessian and sparsity. Value is in  ]0, 1[',
            'default': 0.6,
            'range': (0, 1)
        },
        'reg': {
            'type': float,
            'label': 'Regularization',
            'help': 'Regularization weight. Value is in [0, 1]',
            'default': 0.995,
            'range': (0, 1)
        }
    }
}



"""This module implement utilities for the SAiryscan library"""
from .io import SAiryscanReader
from .settings import Settings, SettingsContainer
from ._observers import SObservable, SObserver, SObserverConsole


__all__ = [
    'SAiryscanReader',
    'Settings',
    'SettingsContainer',
    'SObservable',
    'SObserver',
    'SObserverConsole']



"""This module implements the pseudo-confocal with denoising reconstruction method"""
import torch
from .interface import SAiryscanReconstruction


class PseudoConfocal(SAiryscanReconstruction):
    """Reconstruct an image equivalent to a confocal image by summing the inner detectors

    :param pinhole: Size of the pinhole in AU: [0.6, 1, 1.25]
    """
    def __init__(self, pinhole=1):
        super().__init__()
        self.num_args = 1
        self.pinhole = pinhole
        if pinhole not in [0.6, 1, 1.25]:
            raise ValueError('PseudoConfocal pinhole parameter must be a value in (0.6, 1, 1.25)')

    def __call__(self, image: torch.Tensor) -> torch.Tensor:
        """Do the reconstruction

        :param image: Raw detector stack to reconstruct [H (Z) Y X]
        :return: high resolution image [(Z) Y X]
        """
        self.progress(0)
        out = None
        if self.pinhole == 0.6:
            out = torch.sum(image[0:7, ...], axis=0)
        elif self.pinhole == 1:
            out = torch.sum(image[0:19, ...], axis=0)
        elif self.pinhole == 1.25:
            out = torch.sum(image[0:32, ...], axis=0)
        self.progress(100)
        return self._crop(out)


metadata = {
    'name': 'PseudoConfocal',
    'label': 'Pseudo Confocal',
    'class': PseudoConfocal,
    'parameters': {
        'pinhole': {
            'type': 'select',
            'label': 'Pinhole',
            'values': [0.6, 1, 1.25],
            'help': 'Size of the pinhole (0.6, 1, 1.25)',
            'default': 0.6
        }
    }
}



"""Registration module"""
from .interface import SAiryscanRegistration
from .position import SRegisterPosition
from .fourier_phase import SRegisterFourierPhase
from .mse_registration import SRegisterMSE

from .position import metadata as position_metadata
from .mse_registration import metadata as mse_registration_metadata


metadata = [position_metadata, mse_registration_metadata]

__all__ = [
    'SAiryscanRegistration',
    'SRegisterPosition',
    'SRegisterFourierPhase',
    'SRegisterMSE',
    'metadata'
]



"""This module implements the ISM with denoising reconstruction method"""
import torch
from .interface import SAiryscanReconstruction


class ISM(SAiryscanReconstruction):
    """Reconstruct a high resolution image by summing the co-registered detectors: ISM method"""
    def __init__(self):
        super().__init__()
        self.num_args = 1

    def __call__(self, image: torch.Tensor) -> torch.Tensor:
        """Do the reconstruction

        :param image: Raw airyscan image. [H, Z, Y, X] for 3D image, [H, Y, X] for 2D images
        :return: Tensor: the reconstructed image. [Z, Y, X] for 3D, [Y, X] for 2D

        """
        self.progress(0)
        out = torch.sum(image, axis=0)
        self.progress(100)
        return self._crop(out)


metadata = {
    'name': 'ISM',
    'label': 'ISM',
    'class': ISM,
    'parameters': {
        }
}



"""Implementation of detector registration using the MSE loss"""
import numpy as np
import torch
from torchvision import transforms
from skimage.transform import rescale

from .interface import SAiryscanRegistration


class SRegisterMSE(SAiryscanRegistration):
    """Register the detectors stack by translating each image to the detector position is array

    :param weight: Weight applied on the detector position translation
    """
    def __init__(self, weight: float = 1.0):
        super().__init__()
        self.weight = weight

    def __call__(self, image: torch.Tensor) -> torch.Tensor:
        """DO the registration

        :param image: Single channel/frame stack to register,
        :return: The registered stack
        """
        self.progress(0)
        image_out = torch.zeros(image.shape, dtype=torch.float32)
        image_out[0, ...] = image[0, ...].clone()

        ref_image = rescale(image[0, ...].detach().numpy(), 2)
        # ref_image = ref_image/np.max(ref_image)
        for d in range(1, 32):
            self.progress(int(100*d/32))
            mse = np.zeros((13, 13))
            mov_image_ref = rescale(image[d, ...].detach().numpy(), 2)
            # mov_image_ref = mov_image_ref/np.max(mov_image_ref)
            for x in range(-6, 7):
                for y in range(-6, 7):
                    mov_image = np.roll(mov_image_ref, x, axis=0)
                    mov_image = np.roll(mov_image, y, axis=1)
                    mse[x+5, y+5] = np.mean((ref_image[10:-10, 10:-10] -
                                             mov_image[10:-10, 10:-10])**2)
            min_pos = np.where(mse == np.min(mse))
            shift = ((min_pos[1][0]-5)/2, (min_pos[0][0]-5)/2)
            self.notify(f'shift {d} = ({shift[0]}, {shift[1]})')
            image_out[d, ...] = self._translate_detector(image[d, ...],
                                                         (shift[0] / 2, shift[1] / 2))
            self.progress(100)
        return image_out

    @staticmethod
    def _translate_detector(img, translate):
        """Translate one detector

        :param img: Tensor for one single detector
        :param translate: (x, y, z) or (x, y) translation vector
        :return: the co-registered tensor
        """
        img_bc = img.view((1, 1, img.shape[0], img.shape[1]))
        img_out = transforms.functional.affine(img_bc, angle=0, translate=translate, scale=1,
                        shear=[0, 0],
                        interpolation=transforms.functional.InterpolationMode.BILINEAR)
        return img_out.view(img.shape)


metadata = {
    'name': 'SRegisterMSE',
    'label': 'Translation MSE',
    'class': SRegisterMSE,
    'parameters': {
    }
}
